﻿#pragma TextEncoding = "UTF-8"
#pragma rtGlobals=3
#pragma DefaultTab={3,20,4}
#pragma ModuleName=Hyperspec
#pragma IgorVersion=8
#pragma version=1.01

#include <Resize Controls>

// Project Updater header
static constant kProjectID = 22408 // the project node on IgorExchange
static strconstant ksShortTitle = "Hyperspec" // the project short title on IgorExchange

//#define dynamic

static constant kFTIR = 1
static constant kReverseImageY = 0

menu "New"
	"Hyperspectral browser...", /Q, Hyperspec#NewFPAbrowser()
end

#if IgorVersion() >= 9
#ifdef dynamic
menu "DataBrowserObjectsPopup", dynamic
	Hyperspec#ValidBrowserWave("Hyperspectral browser", 1, 1), /Q, Hyperspec#PlotDataBrowserSelection()
end
#else
menu "DataBrowserObjectsPopup"
	"Hyperspectral browser", /Q, Hyperspec#PlotDataBrowserSelection()
end
#endif
#endif

static function demo()
	make/O/N=(64,64,501) wDemo3d
	// make a broad Gaussian peak that increases in intensity towards the centre of the field of view
	wDemo3d  = 1000 * sqrt((x-32)^2 + (y-32)^2) * gauss(z,350,400)
	// add a linear baseline whose slope varies depending on position
	wDemo3d += (0.0002 * x + 0.0005 * y) * z 
	// add a peak at position 200
	wDemo3d += 500 * gauss(z,200,20)
	// add a peak at position 150 to spectra that fall within a circular area
	wDemo3d += sqrt((x-20)^2 + (y-20)^2) < 10 ? 1000 * gauss(z,150,20) : 0
	// add some noise
	wDemo3d += gnoise(1) 
	MakeFPAbrowser(wDemo3d)
end

static function/S ValidBrowserWave(string str, int selectOne, int FPA)
	if (strlen(GetBrowserSelection(-1)) == 0 || (strlen(GetBrowserSelection(1)) && selectOne))
		return "" // Data Browser is not open/multiple objects selected
	endif
	wave/Z w1 = $GetBrowserSelection(0)
	if (WaveExists(w1) && WaveType(w1,1)==1)
		if (FPA==0)
			return str
		elseif (WaveDims(w1)==3)
			return str
		endif
	endif
	return ""
end

static function Valid3DWave(wave w)
	return  WaveExists(w) && WaveType(w,1)==1 && WaveDims(w)==3
end

static function PlotDataBrowserSelection()
	wave/Z w = $GetBrowserSelection(0)
	if (WaveExists(w) && WaveDims(w)==3)
		MakeFPAbrowser(w)
	endif
end

static function/DF getDFR()
	DFREF dfr = root:Packages:Hyperspec
	if (DataFolderRefStatus(dfr) != 1)
		return resetDFR()
	endif
	return dfr
end

static function/DF resetDFR()
	NewDataFolder/O root:Packages
	NewDataFolder/O root:Packages:Hyperspec
	DFREF dfr=root:Packages:Hyperspec
	return dfr
end

static function NewFPAbrowser()
	CreateBrowser/M Prompt="Select a 3D wave that represents a hyperspectral datacube"
	ModifyBrowser/M showWaves=1, showVars=0, showStrs=0, ShowInfo=0, showPlot=1
	ModifyBrowser/M showModalBrowser
	if (ItemsInList(S_BrowserList) == 1)
		wave/Z wSel = $StringFromList(0, S_BrowserList)
		MakeFPAbrowser(wSel)
	endif
	return 0
end

static function MakeFPAbrowser(wave/Z wFPA)
	
	if (!Valid3DWave(wFPA))
		NewFPAbrowser()
		return 0
	endif
	
	DFREF dfr = GetDFR()
	Make/O/N=(DimSize(wFPA, 0),DimSize(wFPA, 1)) dfr:wMask
	Make/O/N=(DimSize(wFPA, 0),DimSize(wFPA, 1),4) dfr:wDisplayMask/wave=wDisplayMask
	CopyScales wFPA, wDisplayMask
	
	Make/O/N=(DimSize(wFPA, 2)) dfr:spectrum/wave=spectrum
	SetScale/P x, DimOffset(wFPA, 2), DimDelta(wFPA, 2), spectrum
	wDisplayMask[][][0,2] = 65535
	wDisplayMask[][][3] = 0
	
	variable left, top, width, height
	
	GetWindow/Z FPAbrowser wsizeRM
	if (V_flag == 0)
		left = v_left
		top = v_top
	else
		GetMouse
		left = v_left + 100
		top = v_top - 100
	endif

	width = 385
	height = 675
	
	KillWindow/Z FPAbrowser
	NewPanel/W=(left,top,left+width,top+height)/K=1/N=FPAbrowser as "Hyperspectral Browser"
	DoIgorMenu "control", "retrieve window"
	ModifyPanel/W=FPAbrowser cbRGB=(65535,65535,65534)
		
	int xpnt, ypnt, plane
	xpnt = DimSize(wFPA, 0)/2
	ypnt = DimSize(wFPA, 1)/2
	plane = DimSize(wFPA, 2)/2
	spectrum = wFPA[xpnt][ypnt][p]
	string strTextBox = ""
	
	SetVariable svFPAwave,win=FPAbrowser,pos={5,5},size={350,14},title="",value=_STR:GetWavesDataFolder(wFPA, 2), Proc=Hyperspec#SetvariableProc
	SetVariable svFPAwave,win=FPAbrowser, help={"Select a 3D wave that represents a hyperspectral datacube"}
	SetVariable svFPAwave,win=FPAbrowser, userdata=GetWavesDataFolder(wFPA, 2)
	
	Button btnInfo, win=FPAbrowser, pos={365,5}, size={15,15}, Proc=hyperspec#ButtonProc, title=""
	Button btnInfo, win=FPAbrowser, help={"Show Info"}, Picture=hyperspec#pInfo
	
	DefineGuide/W=FPAbrowser FGT={FT,30}, FGM={FT,0.27,FB},FGB={FB,-150}
	Display/W=(21,178,447,433)/FG=(FL,FGM,FR,FGB)/HOST=FPAbrowser/N=G0
	
	AppendImage/W=FPAbrowser#G0 wFPA
	ModifyImage/W=FPAbrowser#G0 ''#0 ctab= {*,*,YellowHot,0}, plane=plane, ctabAutoscale=2
	ModifyGraph/W=FPAbrowser#G0 margin(left)=8,margin(bottom)=1,margin(right)=38
	ModifyGraph/W=FPAbrowser#G0 tick=3,mirror=1,noLabel=2,standoff=0
	AppendImage/W=FPAbrowser#G0 wDisplayMask
	ColorScale/W=FPAbrowser#G0/C/N=scale/F=0/A=RC/X=0/Y=0.00/E/B=1 image=$NameOfWave(wFPA), heightPct=100, width=10, tickLen=4, minor=1
	
	if (kReverseImageY)
		SetAxis/A/R/W=FPAbrowser#G0 left
	endif
	
	// crosshair cursor
	Cursor/W=FPAbrowser#G0/P/I/H=1/C=(65535,65535,65535)/N=1 A $NameOfWave(wFPA) xpnt, ypnt
	sprintf strTextBox, "%s[][][%g]", NameOfWave(wFPA), plane
	TextBox/C/N=text0/F=0/A=LT/X=8/Y=0/E=2 strTextBox
	
	// rather than plotting wFPA[xpnt][ypnt][*] we plot a 1D wave so that it can be easily exported
	Display/W=(39,32,319,116)/FG=(FL,FGT,FR,FGM)/HOST=FPAbrowser/N=G1 spectrum/TN=spectrum
	ModifyGraph/W=FPAbrowser#G1 margin(left)=8,margin(right)=28
	ModifyGraph/W=FPAbrowser#G1 tick(left)=3,noLabel(left)=2,standoff=0,axThick(left)=0
	if (kFTIR)
		SetAxis/A/R/W=FPAbrowser#G1 bottom
	else
		SetAxis/A/W=FPAbrowser#G1 bottom
	endif
	sprintf strTextBox, "%s[%g][%g]", NameOfWave(wFPA), xpnt, ypnt
	TextBox/C/N=text1/F=0/A=LT/X=8/Y=0/E=2 strTextBox
	
	WaveStats/Q/M=1/RMD=[][][plane] wFPA
	variable vMin = v_min, vMax = v_max
	ModifyGraph/W=FPAbrowser#G1 zColor(spectrum)={spectrum,vMin,vMax,YellowHot,0}
	
	// vertical hair cursor
	Cursor/W=FPAbrowser#G1/P/F/S=2/H=2/N=1 B spectrum 0.5, 0.5
		
	SetActiveSubwindow ##
	SetWindow FPAbrowser activeChildFrame=0
	SetWindow FPAbrowser hook(hFPA)=Hyperspec#HookFPA
		
	variable x0 = 28, y0 = 507
	
	SetVariable svPlane,win=FPAbrowser,pos={x0, y0},size={195,180}, value=_NUM:IndexToScale(wFPA,plane,2),Proc=Hyperspec#SetvariableProc
	SetVariable svPlane,win=FPAbrowser,fSize=12,limits={0,Inf,DimDelta(wFPA,2)}
	SetVariable svPlane,win=FPAbrowser,userdata="0" // userdata = "1" to flag updates from cursor hook and avoid reentry
	if (kFTIR)
		SetVariable svPlane,win=FPAbrowser,title="Plane (wavenumber)"
	else
		SetVariable svPlane,win=FPAbrowser,title="Layer"
	endif
	
	y0 += 25
	PopupMenu popCT,win=FPAbrowser,pos={x0,y0},size={160,20}, Proc=Hyperspec#PopupProc
	PopupMenu popCT,win=FPAbrowser,mode=3,value=#"\"*COLORTABLEPOPNONAMES*\"", bodyWidth=160
	CheckBox chkLog,pos={200,y0},size={42,16},title="log scale",fSize=12,value=0,Proc=Hyperspec#checkBoxProc
	
	y0 += 20
	CheckBox chkAutoMin,pos={200,y0},size={42,16},title="auto",fSize=12,value=1,Proc=Hyperspec#checkBoxProc
	SetVariable svMin,pos={32,y0},size={160,18},bodyWidth=70,value=_NUM:vMin, disable=2
	SetVariable svMin,title="First color index",fSize=12,limits={-Inf,Inf,0},Proc=Hyperspec#SetvariableProc
	
	y0 += 20
	Slider sliderMin pos={x0,y0},size={230,30}, vert=0,side=0,limits={vMin,vMax,0},ticks=0,value=vMin, disable=2, Proc=Hyperspec#SliderProc
	
	y0 += 20
	CheckBox chkAutoMax,pos={200,y0},size={42,16},title="auto",fSize=12,value=1,Proc=Hyperspec#checkBoxProc
	SetVariable svMax,pos={32,y0},size={160,18},bodyWidth=70,value=_NUM:vMax, disable=2
	SetVariable svMax,title="Last color index ",fSize=12,limits={-Inf,Inf,0},Proc=Hyperspec#SetvariableProc
	
	y0 += 20
	Slider sliderMax pos={x0,y0},size={230,30}, vert=0,side=0,limits={vMin,vMax,0},ticks=0,value=vMax, disable=2, Proc=Hyperspec#SliderProc
		
	y0 = 507
	x0 = 274
	GroupBox grpPixel,pos={x0,y0},size={86,125},title="Crosshair",fsize=12
	y0 += 22
	x0 += 17
	PopupMenu popColor,win=FPAbrowser,pos={x0,y0},size={50,20},mode=3,value=#"\"*COLORPOP*\"", Proc=Hyperspec#PopupProc
	PopupMenu popColor,win=FPAbrowser,title="",fsize=12,popColor=(65535,65535,65535)
	y0 += 22
	Button btnZap,pos={x0,y0},size={50.00,22.00},Proc=Hyperspec#buttonproc
	Button btnZap,title="Zap"
	y0 += 25
	Button btnCopy,pos={x0,y0},size={50.00,22.00},Proc=Hyperspec#buttonproc
	Button btnCopy,title="Copy"
	y0 += 25
	Button btnPaste,pos={x0,y0},size={50.00,22.00},Proc=Hyperspec#buttonproc
	Button btnPaste,title="Paste"
	
	SetVariable svFPAwave,userdata(ResizeControlsInfo)=A"!!,?X!!#9W!!#Bi!!#;mz!!#](Aon\"Qzzzzzzzzzzzzzz!!#o2B4uAezz"
	SetVariable svFPAwave,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#u:Du]k<zzzzzzzzzzz"
	SetVariable svFPAwave,userdata(ResizeControlsInfo)+=A"zzz!!#u:Du]k<zzzzzzzzzzzzzz!!!"
	Button btnInfo,userdata(ResizeControlsInfo)=A"!!,HqJ,hj-!!#<(!!#<(z!!#o2B4uAezzzzzzzzzzzzzz!!#o2B4uAezz"
	Button btnInfo,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#u:Du]k<zzzzzzzzzzz"
	Button btnInfo,userdata(ResizeControlsInfo)+=A"zzz!!#u:Du]k<zzzzzzzzzzzzzz!!!"
	SetVariable svPlane,userdata(ResizeControlsInfo)=A"!!,CD!!#CbJ,hr(!!#<Hz!!#](Aon\"Qzzzzzzzzzzzzzz!!#](Aon\"Qzz"
	SetVariable svPlane,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	SetVariable svPlane,userdata(ResizeControlsInfo)+=A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	PopupMenu popCT,userdata(ResizeControlsInfo)=A"!!,CD!!#Cj!!#A/!!#<Xz!!#](Aon\"Qzzzzzzzzzzzzzz!!#](Aon\"Qzz"
	PopupMenu popCT,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	PopupMenu popCT,userdata(ResizeControlsInfo)+=A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	CheckBox chkLog,userdata(ResizeControlsInfo)=A"!!,GX!!#Cj!!#??!!#<8z!!#](Aon\"Qzzzzzzzzzzzzzz!!#](Aon\"Qzz"
	CheckBox chkLog,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	CheckBox chkLog,userdata(ResizeControlsInfo)+=A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	CheckBox chkAutoMin,userdata(ResizeControlsInfo)=A"!!,GX!!#Co!!#>6!!#<8z!!#](Aon\"Qzzzzzzzzzzzzzz!!#](Aon\"Qzz"
	CheckBox chkAutoMin,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	CheckBox chkAutoMin,userdata(ResizeControlsInfo)+=A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	SetVariable svMin,userdata(ResizeControlsInfo)=A"!!,C<!!#Co!!#A4!!#<Hz!!#](Aon\"Qzzzzzzzzzzzzzz!!#](Aon\"Qzz"
	SetVariable svMin,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	SetVariable svMin,userdata(ResizeControlsInfo)+=A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	CheckBox chkAutoMax,userdata(ResizeControlsInfo)=A"!!,GX!!#D$!!#>6!!#<8z!!#](Aon\"Qzzzzzzzzzzzzzz!!#](Aon\"Qzz"
	CheckBox chkAutoMax,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	CheckBox chkAutoMax,userdata(ResizeControlsInfo)+=A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	SetVariable svMax,userdata(ResizeControlsInfo)=A"!!,C4!!#D$!!#A5!!#<Hz!!#](Aon\"Qzzzzzzzzzzzzzz!!#](Aon\"Qzz"
	SetVariable svMax,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	SetVariable svMax,userdata(ResizeControlsInfo)+=A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	Slider sliderMin,userdata(ResizeControlsInfo)=A"!!,CD!!#Ct!!#Au!!#<Hz!!#](Aon\"Qzzzzzzzzzzzzzz!!#o2B4uAezz"
	Slider sliderMin,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	Slider sliderMin,userdata(ResizeControlsInfo)+=A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	Slider sliderMax,userdata(ResizeControlsInfo)=A"!!,CD!!#D)!!#Au!!#<Hz!!#](Aon\"Qzzzzzzzzzzzzzz!!#o2B4uAezz"
	Slider sliderMax,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	Slider sliderMax,userdata(ResizeControlsInfo)+=A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
		
	GroupBox grpPixel,userdata(ResizeControlsInfo)=A"!!,HD!!#CbJ,hp;!!#@^z!!#o2B4uAezzzzzzzzzzzzzz!!#o2B4uAezz"
	GroupBox grpPixel,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	GroupBox grpPixel,userdata(ResizeControlsInfo)+=A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	PopupMenu popColor,userdata(ResizeControlsInfo)=A"!!,HLJ,ht?5QF,A!!#<Xz!!#o2B4uAezzzzzzzzzzzzzz!!#o2B4uAezz"
	PopupMenu popColor,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	PopupMenu popColor,userdata(ResizeControlsInfo)+=A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	Button btnZap,userdata(ResizeControlsInfo)=A"!!,HLJ,htD^]6\\l!!#<hz!!#o2B4uAezzzzzzzzzzzzzz!!#o2B4uAezz"
	Button btnZap,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	Button btnZap,userdata(ResizeControlsInfo)+=A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	Button btnCopy,userdata(ResizeControlsInfo)=A"!!,HLJ,htK!!#>V!!#<hz!!#o2B4uAezzzzzzzzzzzzzz!!#o2B4uAezz"
	Button btnCopy,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	Button btnCopy,userdata(ResizeControlsInfo)+=A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	Button btnPaste,userdata(ResizeControlsInfo)=A"!!,HLJ,htQ5QF,A!!#<hz!!#o2B4uAezzzzzzzzzzzzzz!!#o2B4uAezz"
	Button btnPaste,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	Button btnPaste,userdata(ResizeControlsInfo)+=A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	
	SetWindow kwTopWin,userdata(ResizeControlsInfo)=A"!!*'\"z!!#C%J,ht\\5QCcazzzzzzzzzzzzzzzzzzzz"
	SetWindow kwTopWin,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzzzzzzzzzzzzzzz"
	SetWindow kwTopWin,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzzzzzzzzz!!!"
	
	// resizing panel hook
	SetWindow FPAbrowser hook(ResizeControls)=ResizeControls#ResizeControlsHook
	
end

function InfoMenu()
	string cmd = "(Version: " + num2str(GetThisVersion()) + ";"
	cmd += "Visit web page;Email the developer;"
	PopupContextualMenu cmd
	if (V_flag == 2)
		BrowseURL /Z "https://www.wavemetrics.com/node/" + num2str(kProjectID)
	elseif (V_flag == 3)
		Make/T/free/N=5 wt = "mail"
		string newline = "\r\n"
		wt[3] = "subject=" + URLEncode(ksShortTitle + " Package for Igor Pro")
		wt[4] = "Type your message here" + newline + newline
		wt[4] += "PROJECT VERSION:" + num2str(GetThisVersion()) + newline
		wt[4] += ReplaceString(";", IgorInfo(3), newline) + newline
		wt[4] = "body=" + URLEncode(wt[4])
		wt[1] = "tony withers"
		wt[2] = "uni-bayreuth de"
		sprintf cmd, "%sto:%s%%%d%s?%s&%s" wt[0], wt[1], 40, wt[2], wt[3], wt[4]
		BrowseURL/Z ReplaceString(" ", cmd, ".")
	endif
end

menu "GraphMarquee"
	MarqueeMenu("Save Average..."), /Q, Hyperspec#SaveAverage()
end

function/S MarqueeMenu(string s)
	GetMarquee/Z
	if (cmpstr(S_marqueeWin, "FPAbrowser#G0"))
		return ""
	endif
	return s
end

static function SaveAverage()
	DFREF dfr = getDFR()
	wave w = dfr:wAv
	SaveWaveUI(w, basename=NameOfWave(GetImageWave()) + "_av")
end

// display a spectrum representing average of selected pixels
static function AveragePixelsInRect(STRUCT Rect &pix)
	wave wFPA = GetImageWave()
	Duplicate/free/RMD=[pix.left,pix.right][pix.bottom,pix.top][] wFPA wROI
	MatrixOP/free w = sumRows(sumCols(wROI)/numCols(wROI))/numRows(wROI)
	
	wave/SDFR=GetDFR() spectrum
	spectrum = w[0][0][p]
	
	string strTextBox = ""
	sprintf strTextBox, "%s[%g,%g][%g,%g]", NameOfWave(wFPA), pix.left, pix.right, pix.bottom, pix.top
	TextBox/C/W=FPAbrowser#G1/N=text1 strTextBox
end

static function hookFPA(STRUCT WMWinHookStruct &s)
	
	if (s.eventcode == 7) // cursor
		strswitch(s.winName)
		 case "FPAbrowser#G0" : // cursor A in image plot
		 	if (cmpstr(s.cursorname, "A"))
		 		break
		 	endif
		 	// plot the spectrum for the pixel under the cursor
		 	ResetCrosshairSpectrum()
		 	break
		 case "FPAbrowser#G1" : // cursor B in spectrum plot
		 	if (cmpstr(s.cursorname, "B"))
		 		break
		 	endif
		 	// sync setvar with cursor
		 	variable xval = hcsr($s.cursorname, s.winName)
			SetVariable svPlane win=FPAbrowser, value=_NUM:xval // this doesn't trigger update event in setvar proc
			ResetPlots(3)
		 	break
		 endswitch
	endif
	
	GetWindow/Z FPAbrowser hook(hPixels)
	if (strlen(S_Value))
		return 0
	endif
	
	if (s.eventcode == 5 && cmpstr(s.winname, "FPAbrowser#G0") == 0) // mouseup
		STRUCT Rect MQPixels
		if (Marquee2pixels(MQPixels))
			// a marquee is present
			AveragePixelsInRect(MQPixels) // display a spectrum representing average of selected pixels
			DisplayMask(2, pixels=MQPixels)
			ResetPlots(2) // set color as f(z)
		else
			GetWindow/Z FPAbrowser hook(hClosedPath)
			if (strlen(S_Value) == 0)
				DisplayMask(0)
			endif
			ResetCrosshairSpectrum() // resets spectrum IF cursor is on graph; updates plot
		endif
	endif
	
	if (s.eventcode == 11)	// Keyboard event

		if (s.keycode == 65)
			SetAxis/W=FPAbrowser#G0/A
			if (kFTIR)
				SetAxis/A/R/W=FPAbrowser#G1 bottom
			else
				SetAxis/A/W=FPAbrowser#G1 bottom
			endif
			SetAxis/A/W=FPAbrowser#G1 left
			return 1
		endif
		
		if (s.keycode == 90)
			UndoZap()
			return 1
		endif
		
	endif
	
	return 0
end

// plot the spectrum for the pixel under the cursor
static function ResetCrosshairSpectrum()
	if (strlen(CsrInfo(A, "FPAbrowser#G0"))==0)
		ResetPlots(2)
		return 0
	endif
	variable pp = pcsr(A, "FPAbrowser#G0")
	variable qq = qcsr(A, "FPAbrowser#G0")
	wave wFPA = GetImageWave()
	wave/SDFR=GetDFR() spectrum
	spectrum = wFPA[pp][qq][p]
	string strTextBox = ""
	if (hcsr(A, "FPAbrowser#G0")!=pp || vcsr(A, "FPAbrowser#G0")!=qq)
		sprintf strTextBox, "%s[%g][%g] x=%g y=%g", NameOfWave(wFPA), pp, qq, hcsr(A, "FPAbrowser#G0"), vcsr(A, "FPAbrowser#G0")
	else
		sprintf strTextBox, "%s[%g][%g]", NameOfWave(wFPA), pp, qq
	endif
	TextBox/C/W=FPAbrowser#G1/N=text1 strTextBox
	SetWindow FPAbrowser#G0 userdata(hpixel)=num2str(pp)
	SetWindow FPAbrowser#G0 userdata(vpixel)=num2str(qq)
	ResetPlots(2)
	return 1
end

static function Marquee2pixels(STRUCT Rect &MQPixels)
	GetMarquee/Z/W=FPAbrowser#G0 left, bottom
	if (V_flag)
		wave wFPA = GetImageWave()
		int p1, p2, q1, q2, pmin, pmax, qmin, qmax
		q1 = scaleToIndex(wFPA, v_top, 1)
		q2 = scaleToIndex(wFPA, v_bottom, 1)
		p1 = scaleToIndex(wFPA, v_left, 0)
		p2 = scaleToIndex(wFPA, v_right, 0)
		pmin = 0
		qmin = 0
		pmax = DimSize(wFPA, 0) - 1
		qmax = DimSize(wFPA, 1) - 1
		MQPixels.left = max(pmin, min(p1, p2))
		MQPixels.right = min(pmax, max(p1, p2))
		MQPixels.bottom = max(min(q1, q2), qmin)
		MQPixels.top = min(max(q1, q2), qmax)
	endif
	return v_flag
end

static function hookPixelBrush(STRUCT WMWinHookStruct &s)
	
	if (cmpstr(s.winname, "FPAbrowser#G0"))
		return 0
	endif
	
	if (s.eventMod & 16) // allow contextual menu
		return 0
	endif
	
	// need to fix this
	if (s.eventMod & 13) // cmd/ctrl
		s.doSetCursor = 1
//		s.cursorCode = 20, 15, 18
		s.cursorCode = (s.eventMod & 4) ? 19 : 18
	endif
	
	if (s.eventCode==3 && !(s.eventMod&12) )
		PixelEdit(0)
	endif
		
	if (! (s.eventCode==3 && (s.eventMod&12)) ) // mousedown, alt/ctrl/shift-ctrl
		return 0
	endif
	
	wave w_img = GetImageWave()
	int pmax = DimSize(w_img, 0)
	int qmax = DimSize(w_img, 1)
	
	DFREF dfr = GetDFR()
	wave/SDFR=dfr wMask, wDisplayMask
	
	variable v_X, v_Y, p_X, p_Y

	int ctrl = (s.eventMod & 8) > 0
	int opt = (s.eventMod & 4) > 0
	int keys, shift
	keys = ctrl + 2*opt

	int mousebutton = 1
	int oldLeft = s.mouseLoc.h, oldTop = s.mouseLoc.v
	int horiz = 0, vert = 0
	int firstLoop = 1
	for (;mousebutton && (keys&3);)
		keys = GetKeyState(0)
		GetMouse/W=FPAbrowser //$s.WinName
		mousebutton = V_flag & 1
		
		if (keys & 4) // shift
			if ((horiz + vert) == 0)
				if (abs(v_left-oldLeft) > abs(v_top-oldTop))
					vert = 1
				elseif (abs(v_top-oldTop) > abs(v_left-oldLeft))
					horiz = 1
				endif
			endif
			if (vert)
				v_top = oldTop
			elseif (horiz)
				v_left = oldLeft
			endif
		else
			horiz = 0
			vert = 0
		endif
				
		if (v_left==oldLeft && v_top==oldTop && (!firstLoop))
			continue
		endif
		firstLoop = 0
		
		oldLeft = v_left
		oldTop = v_top
		
		ctrl = keys & 1
		opt = keys & 2
		if (!(mousebutton && (ctrl||opt)))
			continue
		endif
		
		int hh = AxisValFromPixel("FPAbrowser#G0", "bottom", v_left)
		int vv = AxisValFromPixel("FPAbrowser#G0", "left", v_top)
		
		int pp = ScaleToIndex(w_img, hh, 0)
		int qq = ScaleToIndex(w_img, vv, 1)
		
		if (pp<0 || qq<0 || pp>=pmax || qq>=qmax)
			continue
		endif
	
		wMask[pp][qq] = ctrl
		wDisplayMask[pp][qq][3] = 40000 * !ctrl
		DoUpdate

	endfor
	AveragePixelsROI(wMask)
	
	return 0
end

static function PopupProc(STRUCT WMPopupAction &s) : PopupMenuControl
	
	if (s.eventCode != 2)
		return 0
	endif
	
	if (cmpstr(s.ctrlname, "popCT") && cmpstr(s.ctrlname, "popColor"))
		return 0
	endif
	
	ResetPlots(3)
	
	return 0
end

static function SetVariableProc(STRUCT WMSetVariableAction &s) : SetVariableControl

	if (s.eventCode<1 || s.eventCode>8)
		return 0
	endif
	
	variable vMin, vMax, vPos
	
	strswitch (s.ctrlName)
		case "svPlane":
			GetAxis/W=FPAbrowser#G1/Q bottom
			Cursor/W=FPAbrowser#G1/F/P B spectrum  (s.dval-V_min)/(V_max-V_min), 0.5
		case "svMin":
		case "svMax": // plane, min or max
			ResetPlots(3)
			break
		case "svFPAwave" :
			if (s.eventcode == 7) // begin edit

					CreateBrowser/M Prompt="Select a 3D wave that represents a hyperspectral datacube"
					ModifyBrowser/M showWaves=1, showVars=0, showStrs=0
					ModifyBrowser/M ShowInfo=0, showPlot=0
					ModifyBrowser/M clearSelection
					if (strlen(s.sval))
						ModifyBrowser/M selectList=s.sval
					endif
					ModifyBrowser/M showModalBrowser
					if (V_Flag == 0)
						// use ControlUpdate to deselect the setvar
						ControlUpdate/W=$s.win $s.ctrlname
						return 0
					endif
					if (ItemsInList(S_BrowserList) == 1)
						wave/Z wSel = $StringFromList(0, S_BrowserList)
						MakeFPAbrowser(wSel)
						ControlUpdate/W=$s.win $s.ctrlname
					endif
			elseif (s.eventcode == 8) // endedit
				wave/Z w = $s.sval
				if (Valid3DWave(w) && (!WaveRefsEqual(w, GetImageWave())))
					MakeFPAbrowser(w)
				else
					SetVariable $s.ctrlName, win=$s.win, value=_STR:GetUserData(s.win, s.ctrlname, "")
				endif
			
			endif
			break
	endswitch
		
	return 0
end

static function ResetSliders([variable limitMin, variable limitMax, variable vMin, variable vMax])
		
	variable vPos
	
	if (ParamIsDefault(vMin))
		ControlInfo/W=FPAbrowser svMin
		vMin = V_Value
	endif
	
	if (ParamIsDefault(vMax))
		ControlInfo/W=FPAbrowser svMax
		vMax = V_Value
	endif
	
	if (ParamIsDefault(limitMin) || ParamIsDefault(limitMax))
		ControlInfo/W=FPAbrowser sliderMin
		vPos = strsearch(s_recreation, "limits=", 0)
		variable v1, v2
		sscanf s_recreation[vPos,Inf], "limits={%g,%g,0}", v1, v2
		limitMin = ParamIsDefault(limitMin) ? v1 : limitMin
		limitMax = ParamIsDefault(limitMax) ? v2 : limitMax
	endif
	
	Slider sliderMin win=FPAbrowser, value=limit(vMin, limitMin, limitMax), limits={limitMin,limitMax,0}
	Slider sliderMax win=FPAbrowser, value=limit(vMax, limitMin, limitMax), limits={limitMin,limitMax,0}
end


function SliderProc(STRUCT WMSliderAction &s) : SliderControl

	if (s.eventcode & 1)
		
		if (cmpstr(s.ctrlname, "sliderMin") == 0)
			SetVariable svMin win=FPAbrowser, value=_NUM:s.curval
		elseif (cmpstr(s.ctrlname, "sliderMax") == 0)
			SetVariable svMax win=FPAbrowser, value=_NUM:s.curval
		endif
		ResetPlots(3)
	endif
	
	if (s.eventmod == 4)
		
	endif
	
	return 0
end


// set plot plane and color table based on control values
static function ResetPlots(int whichone) // bit 0 = image plot, bit 1 = spectrum
	
	int autoMin, autoMax, plane
	variable wavenumber
	variable vMin, vMax, planeMin, planeMax
		
	ControlInfo/W=FPAbrowser chkAutoMin
	autoMin = V_Value
	ControlInfo/W=FPAbrowser chkAutoMax
	autoMax = V_Value
		
	string strImage = StringFromList(0, ImageNameList("FPAbrowser#G0", ";"))
	wave wFPA = ImageNameToWaveRef("FPAbrowser#G0", strImage)
	
	ControlInfo/W=FPAbrowser svPlane
	
	plane = scaleToIndex(wFPA, v_value, 2)
	wavenumber = IndexToScale(wFPA, plane, 2)
	
	WaveStats/Q/M=1/RMD=[][][plane] wFPA
	planeMin = v_min
	planeMax = v_max
		
	if (autoMin)
		vMin= planeMin
		SetVariable svMin win=FPAbrowser, value=_NUM:vMin
	else
		ControlInfo/W=FPAbrowser svMin
		vMin = v_value
	endif
	if (autoMax)
		vMax = planeMax
		SetVariable svMax win=FPAbrowser, value=_NUM:vMax
	else
		ControlInfo/W=FPAbrowser svMax
		vMax = v_value
	endif
	
	ControlInfo/W=FPAbrowser popCT
	string Table = s_value
	
	if (whichone & 1)
		ModifyImage/W=FPAbrowser#G0 ''#0 plane=plane, ctab={vMin,vMax,$Table,0}
		ResetPixelTextbox()

		ControlInfo/W=FPAbrowser PopColor
		variable red, green, blue
		sscanf s_value, "(%d,%d,%d)", red, green, blue
		Cursor/M/W=FPAbrowser#G0/C=(red, green, blue) A

		if (WaveExists(TraceNameToWaveRef("FPAbrowser#G0", "ROI_nodes")))
			ModifyGraph/Z rgb(ROI_nodes)=(red, green, blue)
			wave/Z/SDFR=GetDFR() ROI_x, ROI_y
			Wave2Poly(ROI_x, ROI_y, "bottom", "left", layer="ProgFront", name="ROI")
		endif

		string strTextBox = ""
		sprintf strTextBox, "%s[][][%g]", NameOfWave(wFPA), plane
		TextBox/C/W=FPAbrowser#G0/N=text0 strTextBox
	endif
	
	if (whichone & 2)
		wave spectrum = TraceNameToWaveRef("FPAbrowser#G1", "spectrum")
		ModifyGraph/W=FPAbrowser#G1 zColor(spectrum)={spectrum,vMin,vMax,$Table,0}
		ResetSliders(limitMin=planeMin, limitMax=planeMax, vMin=vMin, vMax=vMax)
	endif
		
	SetVariable svPlane, win=FPAbrowser, value=_NUM:wavenumber
		
end

// returns position (proportion of axis) at axis value = val
static function AxisValToPosition(variable val)
	GetAxis/W=FPAbrowser#G1/Q bottom
	return (val-V_min)/(V_max-V_min)
end

static function CheckBoxProc(STRUCT WMCheckBoxAction &s) : CheckboxControl
	
	if (s.eventCode != 2)
		return 0
	endif
	
	strswitch (s.ctrlName)
		case "chkAutoMin":
			SetVariable svMin  win=FPAbrowser, disable=2*s.checked
			Slider sliderMin   win=FPAbrowser, disable=2*s.checked
			if (s.checked)
				ResetPlots(3)
			endif
			break
		case "chkAutoMax":
			SetVariable svMax  win=FPAbrowser, disable=2*s.checked
			Slider sliderMax   win=FPAbrowser, disable=2*s.checked
			if (s.checked)
				ResetPlots(3)
			endif
			break
		case "chkLog" :
			string strImage = StringFromList(0,ImageNameList("FPAbrowser#G0", ";"))
			ModifyImage/W=FPAbrowser#G0 $strImage log=s.checked
			ColorScale/W=FPAbrowser#G0/C/N=scale log=s.checked
			break
	endswitch
		
	return 0
end

static function ButtonProc(STRUCT WMButtonAction &s) : ButtonControl
	if (s.eventCode != 2)
		return 0
	endif
	
	strswitch (s.ctrlName)
		case "btnZap":
			ZapPixel()
			break
		case "btnCopy":
			CopyPastePixel(0)
			break
		case "btnPaste":
			CopyPastePixel(1)
			break
		case "btnInfo":
			InfoMenu()
			break
	endswitch
	
	return 0
end

static function CopyPastePixel(int paste)
	DFREF dfr = GetDFR()
	wave/Z pixelcopy = dfr:pixelcopy
	wave/Z wFPA = GetImageWave()
	variable xx = hcsr(A, "FPAbrowser#G0")
	variable yy = vcsr(A, "FPAbrowser#G0")
	
	if (paste)
		if (!WaveExists(pixelcopy) || numpnts(pixelcopy)!=DimSize(wFPA, 2))
			return 0
		endif
		
		Duplicate/O/RMD=[xx][yy][] wFPA dfr:backup/wave=bup
		string strNote = ""
		sprintf strNote, "%s;%g;%g;", NameOfWave(wFPA), xx, yy
		note/K bup strNote
		
		wFPA[xx][yy][] = pixelcopy[r]
	else
		Duplicate/O/RMD=[xx][yy][] wFPA dfr:pixelcopy
	endif
end

static function ZapPixel()
	
	// better to check wavenote?
	string str = ""
	#ifdef WINDOWS
	str = "Note: pixel-zapping is for ratioed data!\r\rDo you want to continue?\r\rCtrl-Z to undo."
	#else
	str = "Note: pixel-zapping is for ratioed data!\r\rDo you want to continue?\r\rCmd-Z to undo."
	#endif
	DoAlert 1, str
	if (v_flag == 2)
		return 0
	endif
	
	variable xx = hcsr(A, "FPAbrowser#G0")
	variable yy = vcsr(A, "FPAbrowser#G0")

	wave/Z wFPA = GetImageWave()
	
	if (numtype(xx) || numtype(yy) || !WaveExists(wFPA))
		return 0
	endif
	
	variable i, j
	Make/free/N=(DimSize(wFPA, 2)) total
	variable count
		
	variable newX, newY
	variable xMax = DimSize(wFPA, 0)
	variable yMax = DimSize(wFPA, 1)
	
	for (i=-1;i<2;i++)
		for (j=-1;j<2;j++)
			newX = xx + i
			newY = yy + j
			if (newX>=0 && newY>=0 && newX<xMax && newY<Ymax && (i || j))
				count += 1
				total += wFPA[newX][newY][p]
			endif
		endfor
	endfor
	total /= count
	DFREF dfr = GetDFR()
	Duplicate/O total dfr:backup/wave=bup
	bup = wFPA[xx][yy][p]
	string strNote = ""
	sprintf strNote, "%s;%g;%g;", NameOfWave(wFPA), xx, yy
	note/K bup strNote
	wFPA[xx][yy][] = total[r]
	ResetCrosshairSpectrum()
end

static function/wave GetImageWave()
	string strImage = StringFromList(0, ImageNameList("FPAbrowser#G0", ";"))
	wave/Z wFPA = ImageNameToWaveRef("FPAbrowser#G0", strImage)
	return wFPA
end

static function UndoZap()
	wave/Z wFPA = GetImageWave()
	wave/SDFR=GetDFR() backup
	if (!WaveExists(backup) || !WaveExists(wFPA))
		return 0
	endif
	
	string strNote = note(backup)
	if (cmpstr(NameOfWave(wFPA), StringFromList(0, strNote)))
		return 0
	endif
	
	variable xx = str2num(StringFromList(1, strNote))
	variable yy = str2num(StringFromList(2, strNote))
	
	wFPA[xx][yy][] = backup[r]
	ResetCrosshairSpectrum()
end


// *** save wave ***

static function/WAVE SaveWaveUI(wave w, [string basename, int overwrite])
	
	overwrite = ParamIsDefault(overwrite) ? 1 : overwrite // default is to allow user to overwrite a wave
	if (ParamIsDefault(basename))
		basename = NameOfWave(w)
		if (cmpstr(basename, "_free_") == 0)
			basename = "wave"
		endif
	endif
	KillWaves/Z wwSavedWaves
	DoWindow/K SaveWaveDialog
	GetMouse
	variable left, top, right, bottom
	left = pixel2cpu(V_left) - 50
	top = pixel2cpu(V_top) - 50
	right = left + 220
	bottom = top + 150
	NewPanel/K=1/W=(left,top,right,bottom)/N=SaveWaveDialog as "Save Wave"
	SetVariable svName,pos={25,20},size={157,14},title="Wave Name", userdata(base) = basename
	SetVariable svName,value=_STR:UniqueName(basename, 1, 0), fsize=12, Proc=Hyperspec#SaveWaveSetVars
	PopupMenu popFolder,pos={25,50},size={157,14},title="Data Folder", fsize=12
	PopupMenu popFolder,value=#"GetDataFolder(0)", userdata(folder)=GetDataFolder(1), Proc=Hyperspec#SaveWavePopups
	Button btnCancel,pos={40,100},size={55,20},title="Cancel", Proc=Hyperspec#SaveWaveButtons
	Button btnDoIt,pos={130,100},size={50,20},title="Do It", Proc=Hyperspec#SaveWaveButtons, userdata=num2str(overwrite)

	SetVariable svName,userdata(ResizeControlsInfo)=A"!!,C,!!#<X!!#A,!!#<Hz!!#](Aon\"Qzzzzzzzzzzzzzz!!#o2B4uAezz"
	SetVariable svName,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#u:Du]k<zzzzzzzzzzz"
	SetVariable svName,userdata(ResizeControlsInfo)+=A"zzz!!#u:Du]k<zzzzzzzzzzzzzz!!!"
	PopupMenu popFolder,userdata(ResizeControlsInfo)=A"!!,C,!!#>V!!#@`!!#<Xz!!#](Aon\"Qzzzzzzzzzzzzzz!!#o2B4uAezz"
	PopupMenu popFolder,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#u:Du]k<zzzzzzzzzzz"
	PopupMenu popFolder,userdata(ResizeControlsInfo)+=A"zzz!!#u:Du]k<zzzzzzzzzzzzzz!!!"
	Button btnCancel,userdata(ResizeControlsInfo)=A"!!,D/!!#@,!!#>j!!#<Xz!!#](Aon\"Qzzzzzzzzzzzzzz!!#](Aon\"Qzz"
	Button btnCancel,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	Button btnCancel,userdata(ResizeControlsInfo)+=A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	Button btnDoIt,userdata(ResizeControlsInfo)=A"!!,Fg!!#@,!!#>V!!#<Xz!!#o2B4uAezzzzzzzzzzzzzz!!#o2B4uAezz"
	Button btnDoIt,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	Button btnDoIt,userdata(ResizeControlsInfo)+=A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	SetWindow kwTopWin,userdata(ResizeControlsInfo)=A"!!*'\"z!!#Ak!!#A%zzzzzzzzzzzzzzzzzzzzz"
	SetWindow kwTopWin,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzzzzzzzzzzzzzzz"
	SetWindow kwTopWin,userdata(ResizeControlsInfo)+=A"zzzzzzzzzzzzzzzzzzz!!!"
	
	// resizing panel hook
	SetWindow SaveWaveDialog hook(ResizeControls)=ResizeControls#ResizeControlsHook

	PauseForUser SaveWaveDialog
	wave/Z/WAVE wwSavedWaves
	if (WaveExists(wwSavedWaves))
		wave/Z saved = wwSavedWaves[0]
		string strPathToNewWave = GetWavesDataFolder(saved, 2)
		KillWaves/Z saved
		Duplicate/O w $strPathToNewWave /wave=saved
		KillWaves wwSavedWaves
		return saved
	endif
	return $""
end

static function SaveWavePopups(STRUCT WMPopupAction &s) : PopupMenuControl
	
	if (s.eventCode != 3)
		return 0
	endif

	string strFolder = StringFromList(0, SelectFolder(s.popStr))
	if (strlen(strFolder))
		PopupMenu popFolder, win=$s.win, userdata(folder)=strFolder
		PopupMenu popFolder, win=$s.win, value=#"Hyperspec#GetSelectedFolder(0)"
	endif
	
	CheckNameSetvar()

	return 0
end

static function SaveWaveButtons(STRUCT WMButtonAction &s) : ButtonControl

	if (s.eventcode != 2)
		return 0
	endif
		
	strswitch (s.ctrlName)
		case "btnDoIt":
			int overwrite = str2num(s.userdata)
			ControlInfo/W=$s.win svName
			if (!overwrite && exists(GetSelectedFolder(1)+PossiblyQuoteName(s_value)))
				DoAlert 0, s_value + " exists."
				return 0
			endif
			KillWaves/Z $GetSelectedFolder(1)+PossiblyQuoteName(s_value)
			wave/Z w = $GetSelectedFolder(1)+PossiblyQuoteName(s_value)
			if (WaveExists(w))
				DoAlert 0, s_value + " exists and is in use, cannot overwrite."
				return 0
			endif
			Make/O/N=0 $GetSelectedFolder(1)+PossiblyQuoteName(s_value) /wave=w
			Make/O/WAVE wwSavedWaves = {w}
		case "btnCancel" : // Cancel or Do It
			DoWindow/K $s.win
			break
	endswitch
	return 0
end

static function SaveWaveSetVars(STRUCT WMSetVariableAction &s)
	
	if (s.eventCode < 0) // excludes Igor8 eventcodes -2 and -3
		return 0
	endif
	
	if (cmpstr(s.ctrlName, "svName") == 0)
		CheckNameSetvar()
	endif
	return 0
end

static function CheckNameSetvar()
	ControlInfo/W=SaveWaveDialog svName
	if (strlen(S_Value) == 0)
		S_Value = UniqueName(GetUserData("SaveWaveDialog", "svName", "base"), 1, 0)
		SetVariable svName win=SaveWaveDialog, value=_STR:S_Value
	endif
	if (exists(GetSelectedFolder(1) + S_Value))
		SetVariable svName win=SaveWaveDialog, valueBackColor=(0xFFFF,0,0)
	else
		SetVariable svName win=SaveWaveDialog, valueBackColor=0
	endif
end

static function/T GetSelectedFolder(int fullpath)
	string strFolder = GetUserData("SaveWaveDialog", "popFolder", "folder")
	return SelectString(fullpath, ParseFilePath(0, strFolder, ":", 1, 0), strFolder)
end

static function/T SelectFolder(string selection)
	CreateBrowser/M Prompt="Select Data Folder"
	ModifyBrowser/M showWaves=0, showVars=0, showStrs=0
	ModifyBrowser/M ShowInfo=0, showPlot=0
	ModifyBrowser/M select=selection
	ModifyBrowser/M showModalBrowser
	if (V_Flag == 0)
		return "" // User cancelled
	endif
	return S_BrowserList
end

static function pixel2cpu(variable pixel, [string win])
	variable expansion = ParamIsDefault(win) ? 1 : PanelResolution(win)/PanelResolution("")
	return (ScreenResolution > 96 ? pixel * 72 / ScreenResolution : pixel) / expansion
end

static function/S PolyMenu(string str)
	
	DFREF dfr = GetDFR()
	wave/Z polyX = dfr:ROI_x
	wave/Z polyY = dfr:ROI_y
	
	if (WaveExists(polyY)==0 || numpnts(polyY)<3)
		return ""
	endif
	
	GetLastUserMenuInfo
	if (cmpstr(S_graphName, "FPAbrowser#G0"))
		return ""
	endif
	
	CheckDisplayed/W=FPAbrowser#G0 polyY
	if (!V_flag)
		return ""
	endif
	
	if (abs(polyy[numpnts(polyy)-1]-polyy[0]) > 1e-5)
		return ""
	elseif (abs(polyx[numpnts(polyx)-1]-polyx[0]) > 1e-5)
		return ""
	endif
	
	Make/free wx={AxisValFromPixel(S_graphName, "bottom", V_mouseX)}
	Make/free wy={AxisValFromPixel(S_graphName, "left", V_mouseY)}
	FindPointsInPoly wx, wy, polyx, polyy
	wave W_inPoly
	int inPoly = W_inPoly[0]
	KillWaves/Z W_inPoly
	return SelectString(inPoly, "", str)
end

static function inPoly(variable x, variable y)
	DFREF dfr = GetDFR()
	wave wx = dfr:ROI_x
	wave wy = dfr:ROI_y
	Make/free wxp = {x}
	Make/free wyp = {y}
	
	FindPointsInPoly wxp, wyp, wx, wy
	wave W_inPoly
	int inPoly = W_inPoly[0]
	KillWaves/Z W_inPoly
	return inPoly
end

menu "TracePopup", dynamic
	"-"
	Hyperspec#InGraphMenu("Save Spectrum...", "FPAbrowser#G1"), /Q, Hyperspec#SaveSpectrum()
	Hyperspec#HookActiveMenu("Draw ROI", "hClosedPath", "FPAbrowser#G0", 0), /Q, Hyperspec#StartClosedPath("FPAbrowser#G0", "ProgFront", "left", "bottom", "ROI", 1)
	Hyperspec#HookActiveMenu("Stop ROI Drawing", "hClosedPath", "FPAbrowser#G0", 1), /Q, Hyperspec#StopClosedPath("")
	Hyperspec#HookActiveMenu("Select Pixels", "hPixels", "FPAbrowser#G0", 0), /Q, Hyperspec#PixelEdit(1)
	Hyperspec#HookActiveMenu("Stop Select Pixels", "hPixels", "FPAbrowser#G0", 1), /Q, Hyperspec#PixelEdit(0)

	Hyperspec#InGraphMenu("Draw Shapes", "FPAbrowser#G0"), /Q, Hyperspec#StartDrawing()
	Hyperspec#InGraphMenu("ROI from Shapes", "FPAbrowser#G0"), /Q, Hyperspec#AveragePixelsROI($"")
	Hyperspec#InGraphMenu("Clear Shapes", "FPAbrowser#G0"), /Q, Hyperspec#StopClosedPath(""); SetDrawLayer/W=FPAbrowser#G0/K ProgFront //;
end

menu "GraphPopup", dynamic
	"-"
	Hyperspec#InGraphMenu("Save Spectrum...", "FPAbrowser#G1"), /Q, Hyperspec#SaveSpectrum()
end

static function StartDrawing()
	ControlInfo/W=FPAbrowser PopColor
	variable red, green, blue
	sscanf s_value, "(%d,%d,%d)", red, green, blue
	
	ShowTools/A/W=FPAbrowser arrow
	SetDrawLayer/W=FPAbrowser#G0 ProgFront
	SetDrawEnv/W=FPAbrowser#G0 linefgc=(red, green, blue), fillpat=0, xcoord=bottom, ycoord=left	, Save
end

static function/S InGraphMenu(string str, string strGraph)
	GetLastUserMenuInfo
	if (cmpstr(S_graphName, strGraph))
		return ""
	endif
	return str
end

//function/S ShapesMenu(string str)
////	GetLastUserMenuInfo
////	if (cmpstr(S_graphName, strGraph))
////		return ""
////	endif
//	return str
//end

static function/S HookActiveMenu(string str, string hook, string win, int active)
	GetLastUserMenuInfo
	if (cmpstr(S_graphName, win))
		return ""
	endif
	GetWindow/Z $parent(S_graphName) hook($hook)
	return SelectString((strlen(s_value)>0)==active, "", str)
end

static function saveSpectrum()
	wave w = TraceNameToWaveRef("FPAbrowser#G1", "spectrum")
	SaveWaveUI(w)
end

static function ShowCrosshair(int show)
	if (show && strlen(CsrInfo(A, "FPAbrowser#G0"))==0)
		wave wFPA = GetImageWave()
		variable xpnt = str2num(GetUserData("FPAbrowser#G0", "", "hpixel"))
		variable ypnt = str2num(GetUserData("FPAbrowser#G0", "", "vpixel"))
		xpnt = numtype(xpnt) ? DimSize(wFPA, 0)/2 : xpnt
		ypnt = numtype(ypnt) ? DimSize(wFPA, 1)/2 : ypnt
		ControlInfo/W=FPAbrowser PopColor
		variable red, green, blue
		sscanf s_value, "(%d,%d,%d)", red, green, blue
		Cursor/W=FPAbrowser#G0/P/I/H=1/C=(red, green, blue)/N=1 A $NameOfWave(wFPA) xpnt, ypnt
	elseif (!show)
		Cursor/W=FPAbrowser#G0/K A
	endif
end

// show = 1: mask points outside of poly drawing object created by closed path hook
// show = 2: mask points outside of marquee
// show = 3: mask all points
// show = 4: sync mask with M_ROIMask created by ImageGenerateROIMask
// show = 0: clear mask
static function DisplayMask(variable show, [STRUCT Rect &Pixels])
	wave/SDFR=GetDFR() wDisplayMask
	switch (show)
		case 0 :
			wDisplayMask[][][3] = 0
			break
		case 1 :
			wDisplayMask[][][3] = 40000 * !inPoly(p, q)
			break
		case 2 :
			wDisplayMask[][][3] = 40000 * !((p+0.5)==limit(p+0.5, pixels.left, pixels.right) && (q+0.5)==limit((q+0.5), pixels.bottom, pixels.top))
			wDisplayMask[][][3] = 40000 * !((p)==limit(p, pixels.left, pixels.right) && (q)==limit((q), pixels.bottom, pixels.top))
			break
		case 3 :
			wDisplayMask[][][3] = 40000
			break
		case 4 :
			wDisplayMask[][][3] = 40000
			wave/Z M_ROIMask
			if (WaveExists(M_ROIMask))
				wDisplayMask[][][3] = 40000 * !M_ROIMask[p][q]
			endif
			break
		default:
			wDisplayMask[][][3] = 0
	endswitch
	ShowCrosshair(!show)
end

// if wMask is null, an image mask is created from drawing layer object(s)
static function AveragePixelsROI(wave/Z wMask)
		
	if (!WaveExists(wMask))
		KillWaves/Z M_ROIMask
		ImageGenerateROIMask/W=FPAbrowser#G0 $StringFromList(0,ImageNameList("FPAbrowser#G0", ";"))
		if (!v_flag)
			DoAlert 0, "No drawing objects to analyse"
			return 0
		endif
		wave wMask = M_ROIMask
		DisplayMask(4)
	endif
	
	wave wFPA = GetImageWave()
	Duplicate/free wFPA wROI
	wROI = wMask[p][q] ? wROI : 0
	
	MatrixOP/free w = sumRows(sumCols(wROI))
	variable numSpectra = sum(wMask)
	
	wave/SDFR=GetDFR() spectrum
	spectrum = w[0][0][p] / numSpectra
	
	string strTextBox = ""
	sprintf strTextBox, "Masked Region"
	TextBox/C/W=FPAbrowser#G1/N=text1 strTextBox
	
	ResetPlots(2)
	
	KillWaves/Z M_ROIMask
end

static function PixelEdit(int start)
	if (start)
		StopClosedPath("FPAbrowser#G0")
		wave/SDFR=GetDFR() wMask
		wMask = 0
		DisplayMask(3)
		SetWindow FPAbrowser hook(hPixels)=Hyperspec#hookPixelBrush
		ResetPixelTextbox()
		AveragePixelsROI(wMask)
	else
		SetWindow FPAbrowser hook(hPixels)=$""
		ResetPixelTextbox()
		DisplayMask(0)
		ResetCrosshairSpectrum()
	endif
end

static function ResetPixelTextbox()
	GetWindow/Z FPAbrowser hook(hPixels)
	if (strlen(s_value))
		ControlInfo/W=FPAbrowser PopColor
		variable red, green, blue
		sscanf s_value, "(%d,%d,%d)", red, green, blue
		string strTitle = ""
		#ifdef WINDOWS
		sprintf strTitle, "\\K(%d,%d,%d)Ctrl to select, Alt to deselect", red, green, blue
		#else
		sprintf strTitle, "\\K(%d,%d,%d)Cmd to select, Option to deselect", red, green, blue
		#endif
		TextBox/W=FPAbrowser#G0/C/N=textPixels/F=0/A=LT/B=1 strTitle
	else
		TextBox/W=FPAbrowser#G0/K/N=textPixels
	endif
end

// *** draw shape ***

static function StopClosedPath(string strGraph)
	if (strlen(strGraph)==0)
		GetLastUserMenuInfo
		strGraph = S_graphName
	endif
	SetWindow $parent(strGraph) hook(hClosedPath)=$""
	wave/Z/SDFR=getDFR()/WAVE refs
	if (WaveExists(refs))
		RemoveFromGraph/Z/W=$strGraph $NameOfWave(refs[0]), $NameOfWave(refs[2])
	endif
	DrawAction/L=ProgFront getgroup=ROI, delete
	DisplayMask(0)
	
	ResetCrosshairSpectrum()
end
	
static function Wave2Poly(wave/Z wx, wave/Z wy, string hAxis, string vAxis, [string layer, string name])
	
	layer = SelectString(ParamIsDefault(layer), layer, "ProgFront")
	name = SelectString(ParamIsDefault(name), name, "")
	
	if (DimSize(wy, 0)<2)
		return 0
	endif
	
	name = SelectString(strlen(name)==0, name, NameOfWave(wy) + "Poly")
	ControlInfo/W=FPAbrowser PopColor
	variable red, green, blue
	sscanf s_value, "(%d,%d,%d)", red, green, blue
	
	DrawAction/L=$layer getgroup=$name, delete
	SetDrawLayer $layer
	SetDrawEnv gstart, gname=$name
	SetDrawEnv xcoord=$hAxis, ycoord=$vAxis
	SetDrawEnv fillpat= 0, linefgc= (red, green, blue)
	DrawPoly wx[0], wy[0], 1, 1, wx, wy
	SetDrawEnv gstop, gname=$name

	WaveStats/Q/M=1 wy
	return ((V_numNaNs + V_numInfs) == 0)
end


// shape = 1 to create a drawing object.
static function StartClosedPath(string Graph, string layer, string vAxis, string hAxis, string name, int shape)
	
	PixelEdit(0)
	
	DFREF dfr = GetDFR()
	
	Make/O/N=(0,2) dfr:$(name+"_nodes")/wave=nodes
	Make/O/N=(0) dfr:$(name+"_x")/wave=wx
	Make/O/N=(0) dfr:$(name+"_y")/wave=wy
	note wy "layer="+layer+";name="+name+";wx="+GetWavesDataFolder(wx, 2)+";nodes="+GetWavesDataFolder(nodes, 2)+";"
		
	ControlInfo/W=FPAbrowser PopColor
	variable red, green, blue
	sscanf s_value, "(%d,%d,%d)", red, green, blue
			
	RemoveFromGraph/Z $NameOfWave(wy)
	AppendToGraph/W=FPAbrowser#G0 wy vs wx
	ModifyGraph/Z mode($NameOfWave(wy))=0
	
	string strNodes = NameOfWave(nodes)
	RemoveFromGraph/Z $strNodes // keep trace on top
	AppendToGraph/W=FPAbrowser#G0 nodes[][1] vs nodes[][0]
	
	ModifyGraph mode($strNodes)=3, marker($strNodes)=8, rgb($strNodes)=(red, green, blue)
	if (shape) // make spline transparent (not hidden, because we want to be able to right-click and edit nodes)
		ModifyGraph rgb($NameOfWave(wy))=(65535,0,0,0)
	endif
	
	Make/O/WAVE/N=3 dfr:refs/wave=refs
	refs[0] = nodes
	refs[1] = wx
	refs[2] = wy
	
	string/G dfr:win = Graph
	string/G dfr:layer = layer
	variable/G dfr:status = (shape != 0)
	variable/G dfr:pointNum = NaN
	variable/G dfr:mode = 0
	variable/G dfr:success = 1
	
	// clear any other active closed path hook
	wave/T wWins = ListToTextWave(WinList("!"+Graph,";","WIN:65"), ";")
	for (string win : wWins)
		SetWindow $win hook(hClosedPath)=$""
	endfor
	DisplayMask(0)
	
	DoWindow/F $parent(Graph)
	SetActiveSubwindow $Graph
	
	SetWindow $parent(Graph) hook(hClosedPath)=Hyperspec#hookClosedPath
	
	ShowCrosshair(0)
	wave/SDFR=GetDFR() wMask
	wMask = 0
	AveragePixelsROI(wMask)
end

// corrects for aspect ratio of graph area and relative scale of axes.
static function AspectRatio(string strGraph, string hAxis, string vAxis)
	GetAxis/W=$strGraph/Q $hAxis
	variable horizontalAx = V_max - V_min
	
	string strInfo = AxisInfo(strGraph, hAxis)
	if( NumberByKey("log(x)",strInfo,"="))
		horizontalAx = log(V_max) - log(V_min)
	endif
	
	GetAxis/W=$strGraph/Q $vAxis
	variable verticalAx = V_max - V_min
		
	strInfo = AxisInfo(strGraph, vAxis)
	if( NumberByKey("log(x)",strInfo,"="))
		verticalAx = log(V_max) - log(V_min)
	endif
		
	GetWindow $strGraph psize
	return abs(horizontalAx/verticalAx) / abs((V_left - V_right)/(V_top - V_bottom))
end

static function hookClosedPath(STRUCT WMWinHookStruct &s)
	
	DFREF dfr = getDFR()
	SVAR/SDFR=dfr win
	if (cmpstr(s.winname, win))
		return 0
	endif

	// we will use cursor codes to define the mode
	variable Insert    = 18 // insert node
	variable Move      = 21 // move node
	variable Zap       = 19 // delete node
	variable CloseLoop = 33
	variable Idle      = 0
	
	// status definitions
	variable wavedraw  = 0
	variable polydraw  = 1
	variable editnodes = 2
	
	wave/wave/Z refs = dfr:refs
	if (WaveExists(refs) == 0)
		SetWindow $parent(s.winName) hook(hClosedPath)=$""
		return 0
	endif
	NVAR status = dfr:status
	NVAR mode = dfr:mode
	NVAR pointNum = dfr:pointNum
	NVAR polysuccess = dfr:success
	SVAR layer = dfr:layer // when layer is set there is a drawing object associated with the nodes
	
	wave/Z nodes = refs[0]
	wave/Z wx    = refs[1]
	wave/Z wy    = refs[2]
	
	if (!(WaveExists(nodes) && WaveExists(wy) && WaveExists(wx)))
		SetWindow $parent(s.winName) hook(hClosedPath)=$""
		return 0
	endif
	
	string strNodes   = NameOfWave(nodes)
	string strTrace   = NameOfWave(wy)
	string strObjName = StringByKey("name", note(wy), "=") // name for drawing object
	
	string strInfo = TraceInfo(s.winName, strNodes, 0)
	string hAxis   = StringByKey("XAXIS", strInfo)
	string vAxis   = StringByKey("YAXIS", strInfo)
	
	strInfo  = AxisInfo(s.winName, hAxis)
	int logx = NumberByKey("log(x)",strInfo,"=")
	strInfo  = AxisInfo(s.winName, vAxis)
	int logy = NumberByKey("log(x)",strInfo,"=")
		
	switch(s.eventCode)
		case 4: // mousemoved
			if (s.eventmod & 8) // command key
				mode = Zap // delete node
				s.cursorCode = Zap
			elseif (mode == Move) // move node
				s.cursorCode = Move
			else // zap mode will be decided based on cursor position and status
				mode = Idle
				pointNum = NaN
			endif
			
			if (mode == Move) // move node and update spine
				nodes[pointNum][] = {{AxisValFromPixel(s.winName, hAxis, s.mouseloc.h)},{AxisValFromPixel(s.winName, vAxis, s.mouseloc.v)}}
				CatmullRomSpline(nodes, wx, wy, closed=(status==editnodes), aspect=AspectRatio(s.winName, hAxis, vAxis), logx=logx, logy=logy)
				// redraw poly if we move the anchor point
				if (pointnum==0 && strlen(layer))
					Wave2Poly(wx, wy, hAxis, vAxis, layer=layer, name=strObjName)
				endif
			else
				pointNum = NodeUnderCursor(s, nodes)
				
				if (mode != Zap)
					if (numtype(pointNum) == 0)
						if (status == editnodes || pointnum>0 || DimSize(nodes, 0)<=2)
							s.cursorCode = Move // cursor indicates that node can be moved
							mode = Idle // don't enter move mode until mousedown
						elseif (DimSize(nodes, 0) > 2) // ensure that we enclose a finite volume
							s.cursorCode = CloseLoop // cursor indicates that loop can be closed
							mode = CloseLoop
						endif
					else
						pointNum = SegmentUnderCursor(s, wy)
						if (numtype(pointNum) == 0) // cursor over segment
							mode = Insert
						elseif (status != editnodes && MouseInPlotFrame(s, hAxis, vAxis))
							pointNum = DimSize(nodes, 0)
							mode = Insert
						endif

						if (mode == Insert) // in position to insert a point
							mode = Insert
							s.cursorCode = Insert
						endif
					endif
				endif
			endif
			break
		case 3: // mousedown
			
			if (s.eventmod == 17) // right click
				return 0
			endif
			
			if (numtype(pointnum)==0 && mode==Idle) // grab node beneath cursor for moving
				mode = Move
			endif
			
			if (mode == Insert) // insert a node
				InsertPoints/M=0 pointnum, 1, nodes
				nodes[pointnum][0] = AxisValFromPixel(s.winName, hAxis, s.mouseloc.h)
				nodes[pointnum][1] = AxisValFromPixel(s.winName, vAxis, s.mouseloc.v)
				// allow the inserted node to be dragged while the mouse button remains down
				mode = Move
				s.cursorCode = Move
			endif
						
			break
		case 5: // mouseup
			if (mode == Zap && numtype(pointNum) == 0) // delete a node
				DeletePoints/M=0 pointNum, 1, nodes
				
				if (pointnum==0 && strlen(layer))
					polysuccess = 0 // trigger an update of the drawing object
				endif
				
				pointNum = NaN
			elseif (mode == CloseLoop) // status changes from 'draw' to 'edit' after closing the loop
				status = editnodes
				s.cursorCode = Move
			else
				// stop adjusting points after clicking elsewhere in window
				if (mode != Move && status == editnodes)
//					SetWindow $parent(s.winName) hook(hClosedPath)=$""
//					RemoveFromGraph/Z/W=$s.winName $strNodes
					StopClosedPath(s.winname)
					return 0
				endif
				
				// stop moving nodes after mouseup
				mode = Idle

			endif
			
			CatmullRomSpline(nodes, wx, wy, closed=(status==editnodes), aspect=AspectRatio(s.winName, hAxis, vAxis), logx=logx, logy=logy)
			
			if (polysuccess == 0 || (status==polydraw && DimSize(nodes, 0)>1))
				// draw the poly (the drawing object becomes dependent upon wx & wy waves) and revert to wavedraw status on success
				polysuccess = Wave2Poly(wx, wy, hAxis, vAxis, layer=layer, name=strObjName)
				if (polysuccess && (status==polydraw))
					status = wavedraw
				endif
			endif
			
			if (status == editnodes)
				DisplayMask(1)
				AveragePixelsROI($"")
			endif
			
			break
		case 6: //resize
			// even though we're using axis coordinates, if the window aspect ratio changes, the path shape will change
			CatmullRomSpline(nodes, wx, wy, closed=(status==editnodes), aspect=AspectRatio(s.winName, hAxis, vAxis), logx=logx, logy=logy)
			if (status == editnodes)
				DisplayMask(1)
				AveragePixelsROI($"")
			endif
			break
	endswitch
	
	if (s.cursorCode)
		s.doSetCursor = 1
	endif
	
	return 0
end

static function/S parent(string str)
	return ParseFilePath(0, str, "#", 0, 0)
end

static function MouseInPlotFrame(STRUCT WMWinHookStruct &s, string hAxis, string vAxis)
	variable xx = AxisValFromPixel(s.winName, hAxis, s.mouseloc.h)
	variable yy = AxisValFromPixel(s.winName, vAxis, s.mouseloc.v)
	GetAxis/W=$s.winName/Q $hAxis
	variable xmax = max(v_min,v_max)
	variable xmin = min(v_min,v_max)
	GetAxis/W=$s.winName/Q $vAxis
	variable ymax = max(v_min,v_max)
	variable ymin = min(v_min,v_max)
	return (xx == limit(xx, xmin, xmax) && yy == limit(yy, ymin, ymax))
end

static function NodeUnderCursor(STRUCT WMWinHookStruct &s, wave nodes)
	string info = TraceFromPixel(s.mouseloc.h, s.mouseloc.v, "WINDOW:"+s.winName+";DELTAX:4;DELTAY:4;")
	string tname = StringByKey("TRACE", info)
	wave target = TraceNameToWaveRef(s.winName, tname)
	if (WaveRefsEqual(nodes, target))
		return NumberByKey("HITPOINT", info)
	endif
	return NaN
end

static function SegmentUnderCursor(STRUCT WMWinHookStruct &s, wave wy)
	string info = TraceFromPixel(s.mouseloc.h, s.mouseloc.v, "WINDOW:"+s.winName+";DELTAX:4;DELTAY:4;")
	string tname = StringByKey("TRACE", info)
	wave target = TraceNameToWaveRef(s.winName, tname)
	if (WaveRefsEqual(wy, target))
		return ceil(NumberByKey("HITPOINT", info) / 20)
	endif
	return NaN
end

// Create a spline that passes through a series of x-y points,
// optionally looping back to the start point.
// Can be used with ConvexHull to make a loop around a cloud of points.
static function CatmullRomSpline(nodes, wx, wy, [segPoints, alpha, overwrite, closed, name, aspect, logx, logy])
	wave nodes
	wave/Z/D wx, wy
	variable segPoints, alpha
	int overwrite, closed
	string name
	variable aspect, logx, logy
	
	segPoints = ParamIsDefault(segPoints) ? 20 : segPoints
	alpha     = ParamIsDefault(alpha) ? 0.5 : alpha
	overwrite = ParamIsDefault(overwrite) ? 0 : overwrite
	closed    = ParamIsDefault(closed) ? 0 : closed
	name      = SelectString(ParamIsDefault(name), name, NameOfWave(nodes)+"_CR")
	aspect    = ParamIsDefault(aspect) ? 1 : aspect
	logx      = ParamIsDefault(logx) ? 0 : logx
	logy      = ParamIsDefault(logy) ? 0 : logy
		
	variable np = DimSize(nodes, 0)
	
	if (!(WaveExists(wy) && WaveExists(wx)))
		wave/Z/D wx = $name+"_x"
		wave/Z/D wy = $name+"_y"
		if (WaveExists(wx) || WaveExists(wy))
			if (!overwrite)
				DoAlert 1, "overwrite existing wave?"
				if (V_Flag==2)
					return 0
				endif
			endif
			Redimension/N=(0)/D wx, wy
		else
			Make/D/O/N=(0) $name+"_x" /WAVE=wx
			Make/D/O/N=(0) $name+"_y" /WAVE=wy
		endif
	else
		Redimension/N=(0)/D wx, wy
	endif
		
	if (np < 2)
		return 0
	endif
		
	Make/free/N=(np+2,2)/D temp
	temp[1,np][] = nodes[p-1][q]
	
	if (logx == 1)
		temp[1,np][0] = log(temp[p][0])
	elseif (logx == 2) // this is rather pointless. doesn't matter which log base you use.
		temp[1,np][0] = log(temp[p][0])/log(2)
	endif
	if (logy == 1)
		temp[1,np][1] = log(temp[p][1])
	elseif (logy == 2)
		temp[1,np][1] = log(temp[p][1])/log(2)
	endif
	if (aspect != 1)
		temp[1,np][1] *= aspect
	endif
		
	if (closed)
		// check whether nodes form a closed or open loop
		if (nodes[0][0]!=nodes[np-1][0] || nodes[0][1]!=nodes[np-1][1])
			np += 1
			Redimension/N=(np+2,2) temp
			temp[np][] = temp[1][q]
		endif
		
		// we have made sure the nodes wrap around in points 1 ... np,
		// figure out values for points 0 and np+1.
		temp[0][] = temp[np-1][q]
		temp[np+1][] = temp[2][q]
	else // extrapolate beyond first and last node
		temp[0][] = temp[1][q] - (temp[2][q] - temp[1][q])
		temp[np+1][] = temp[np][q] + (temp[np][q] - temp[np-1][q])
	endif
		
	Make/free/N=(segPoints,2) xy_out
	Make/free/N=(4,2) xy_in
	
	variable i
	for (i=1;i<np;i+=1)
		xy_in[][] = temp[i-1+p][q]
		CatmullRomSegment(xy_in, xy_out, alpha)
		SplitWave/free/OREF=refs xy_out
		Concatenate/NP {refs[0]}, wx
		Concatenate/NP {refs[1]}, wy
	endfor

	wx[numpnts(wx)]={temp[np][0]}
	wy[numpnts(wy)]={temp[np][1]}
	
	if (aspect != 1)
		wy /= aspect
	endif
	if (logx == 1)
		wx = 10^wx
	elseif (logx == 2)
		wx = 2^wx
	endif
	if (logy == 1)
		wy = 10^wy
	elseif (logy == 2)
		wy = 2^wy
	endif
end

static function CatmullRomSegment(wave xy_in, wave xy_out, variable alpha)
	// xy_in has x and y values of four control points
	// alpha=0 for standard (uniform) Catmull-Rom spline
	// alpha=0.5 for centripetal Catmull-Rom spline
	// alpha=1 for chordal Catmull-Rom spline
	
	int nPoints = DimSize(xy_out, 0)
	Make/D/free/N=4 T4 = 0
	T4[1,3] = sqrt((xy_in[p][0]-xy_in[p-1][0])^2 + (xy_in[p][1]-xy_in[p-1][1])^2)^alpha + T4[p-1]
	Make/D/free/N=(nPoints, 3, 2) wA
	Make/D/free/N=(nPoints, 2, 2) wB
	Make/D/free/N=(nPoints) w_t
		
	w_t = T4[1] + p/(nPoints)*(T4[2]-T4[1])
	wA = (T4[q+1]-w_t[p]) / (T4[q+1]-T4[q]) * xy_in[q][r] + (w_t[p]-T4[q]) / (T4[q+1]-T4[q]) * xy_in[q+1][r]
	wB = (T4[q+2]-w_t[p]) / (T4[q+2]-T4[q]) * wA[p][q][r] + (w_t[p]-T4[q]) / (T4[q+2]-T4[q]) * wA[p][q+1][r]
	xy_out = (T4[2]-w_t[p])/(T4[2]-T4[1])*wB[p][0][q] + (w_t[p]-T4[1])/(T4[2]-T4[1])*wB[p][1][q]
end

static function GetThisVersion()

	variable procVersion
	// try the quick way
	#if exists("ProcedureVersion") == 3
	procVersion = ProcedureVersion("")
	if (procVersion)
		return procVersion
	endif
	#endif
	
	int maxLines = 30 // number of lines to search for version pragma
	int refNum, i
	string strHeader = ""
	string strLine = ""
	
	Open/R/Z refnum as FunctionPath("")
	if (refnum == 0)
		return 0
	endif
	for (i=0;i<maxLines;i+=1)
		FReadLine refNum, strLine
		strHeader += strLine
	endfor
	Close refnum
	wave/T ProcText = ListToTextWave(strHeader, "\r")
	string S_Value = "" // workaround for Grep bug for Igor 8
	Grep/Q/E="(?i)^#pragma[\s]*version[\s]*="/LIST/Z ProcText
	if (v_flag != 0)
		return 0
	endif
	s_value = LowerStr(TrimString(s_value, 1))
	sscanf s_value, "#pragma version = %f", procVersion
	ProcVersion = (V_flag!=1 || ProcVersion<=0) ? 0 : ProcVersion

	return ProcVersion
end


// PNG: width= 90, height= 30
static Picture pInfo
	ASCII85Begin
	M,6r;%14!\!!!!.8Ou6I!!!"&!!!!?#R18/!3BT8GQ7^D&TgHDFAm*iFE_/6AH5;7DfQssEc39jTBQ
	=U#t@KS5u_NKc'hjQ<7_k>^N6#'iu's#@iL6?EN#,SR%ORqWY?g_DQ6ZakX-9m6UF`Y7B5's^*$E%B
	u3<(a:i;5<Ct\*^a)hT'K%0QC]d[0jPV;N@+9cG@fjB`p38^oR<*NngmMP4QXaLkr6j/Tp8k\$ZiA^
	F[J/@;&;!fa/0Y"8EKlO!*92tV8I!509L_n[prdb[0DD]b^77\4Qh+<X22WSL=^Tpl!>%R?bOV1l`F
	.H/kj.Q0YHO!p0Z_o!(t!"2!0>1<o&\&oX/k93#9I:/\ol*"]HFk2"&^kI,Op^Po$LJ:qE[=Zdn*%1
	/%c)$r:,4%r)#UgVCL9-4fb:)2uc00,SCF2g"FKhP:)ciSp&'<om1=hD5F)*"ddDT8%a$Z"blrKB1V
	t"bqBnEcNrlWWRU];Q?b;\6c%+`!0V&SGg^%NRV$e":,J(6+Q,'.qs8WXZ?]HQd?\@;-q_mAJ,]9,i
	ondb3ItHg(Zjc?GW[$h>t<@#!7L*gWDnag1eK/1,*>Cr1P3*D[BX`3`d.+2D9NUG(..rXX^7$1"@82
	!1KE^Ub4(b^NWMsF;Ht6&&;b$b#dun1L:o\#UFM.3gNK`n[l7Cp+c[X/cDaa?O<b.!b=-Pr2.^?1Nd
	)e2#U:Q=,SCE:&8^%*L$dX(UQj\B/Bea\(..rm00Qspp'-Bu;ci@[g2Ab(7h`\ZA=S45.(4!_jMmH)
	qUNK^MN[#3[D>08n\uMdd,!N;U*ru@<Dd(/5&X;O&;3@n7RotRD4?;%HAGk=G+fW.>tiHTqLp0W7=4
	1t\u)fWAJng`QIdQ"8TJr%$paau+6@''?G#H(RO^<AUs^^G[]\cSB@c5aI;=JigY9^3e#1T9SXPucg
	hq>mG`bOMiSe2FC2.Jjp@_T>`piIOWDnc;k%;&=gQn1M0p,[l:OIIe[!*"QkrNrEAXG*0N(r07]$fZ
	h/$5=fV3Tb,=$\psV/^EJ)4IFB]PEW'\qbg(SsMlA:JXaHWi@OR/f1s5k:MsaPoMc&J--#d&-rCGg1
	@0&:")(CjpXC94`bia)`#WJ(G?agO/a)!%!Za'1b>,JjlN'!=]&V5U;CI@Et9%o2-h+N"!1EAB5`,M
	Th&'lHW;8P&FJlPBUVCZ77C(9D;1qiXUW&&O5KD=a,V0/D%4B*&."qC)Xk?1%Zs:M/7mre"!%<ET2<
	bOD<cB%HGLYk/]p$:.Z,bn[7sj`?B9ZXXV6Rq]_Pti2gd\J3f(i:R&5_pm4KmDXm$aS/B^J!/^)#uG
	%6sfYult1cH[_1nRTfUip]H>d<$0>4r9NH8sWQ9)djkQAg3E#:]C\]o'uTN1rJBf7mE:cKPfJ<US@o
	sfC=1##U)g+pYUIm80gMI'&gcc'97IQ?8%%Ad+-[#D;.]AGGtPU1e&m.iUpnuW^=]Xm<&%d+jRgmR6
	q<!"K4)c%%V>s6.boU<k:W9@PN6o+0)@'G&kQ@q=8),*BN8CpR@_H@2O[8@=!9Ifs"n;A,h79HFe;'
	<AQ<<*SW3\4hYn'7c+!LCK<HgOtdE6\7T^B_C4`os.\g?L1-<OUon"8QFD=fRPcBo,tYs-%!5e<il-
	gK-hM%Q-;cW@=dI0o(g+V`I[7Hll^gmF1C8U`9;n:SDf?'4kXX)YA6CN6o]r]qT!s1+^3pK)B6pt_B
	a+O@`IRg8=:@XkJL"[Wg"k-TK%RqsK*5:)MTX360]U/4Wi@77;Jt3uP7<CJ.#Pn:q"VM]0-o&(8.o9
	`MEo2(5'MS@J+Rq@q<5kKI!A\?`c`\j5Xs.C`^\u^];Gr<!]0NL(T`^;0>7#+XNrGfbj/DFR`<:aKO
	an5L'f'H$C6<N=ieH_Ns:8r/-F>E>*(o2.ET1%BpgNq9:F_:'kIT8!)),C3GlcmYHU;j&0aK<:WMel
	3Ze-Om%!i`\)#_:DlKDdZeWg@P_C[O5M'iUB8>_-2HE;O1>jM;$8=V+SNHPXHb;/UA2:]pC+19D\Bd
	qtPsUXEa5qKga#B2SZ=.SRp%A!RdqL(!Ea7bt3ql@F;Uj=M08??L+[=OW75>HDQ5.GO:69br-505tC
	KaJC@jCMg**JRJQascSIH0iWNK&pUTkRi35A[F<CY2`i#$#_f\:6RcLI9_:V"&,4[]HN!OfQlP*Qtb
	VGkuA?F_!6F<)ocgU#D#V(':e98CFiHpC[p2PjB,D#;1";i6n,<g%nVGOf<>lBjHXs8BA(fD;DdgE&
	_+N0,P*1$jPnaB'W+c:JFR2daE`AbT2\$"\Atc30k=3]2,.R>E*Nf"c^j@=F?-EhKuZgDlI9;bg[c+
	#@$`/2$kB@MhsTDp>G]B0oZ#`%fgeAhHcB?i5($LInM,P9MS?gJ,auQDYK=UG;=]pX[t0`Ddn+h)@5
	*f*'#BK+[7+e+25ad[WOTm!ShS%o^pU.1GgqQgF=j%Y7Q;ZTcp"RlDZ'E)=kG_qYSr#Wb]AY@*-(nJ
	s#A#q"YFMfsgRh]Kt<,gI6=12lg#`?;TLp_*KtthTY]\gR%W:!I8"sroVaPFBo;RFjF5*7K$9'/M"a
	&=8b8<:r],W]a;6"MF%U=g2!m3l-^kmCMU]X>\[CWf<8QErql0Iq<+A?>3\a$'Y&&M_Vt+amGf@imo
	bQW3d]\44-GFK,M+U?MapjQn$rZ2MO]A366&(1N))Y5K\G-`9XF4l;%/:LK#,8h*dQfYG3mlOSIcZr
	>bLn*pC6ZP>8Kg\bG_@CKr2I%"\bg,9T7gnab8=Dm!=-WrDiH(kdbct`7o4FNN0!X<E0c>,]?+,R:L
	D]PZ0,$>?`&92Jr:c#b4?2.q.C==rdOGT-GEe&/PNlY926R6do>OS3)p0O\R3Y8&I`m-S_GCD(TPUp
	#Oh0$f?=CMoA(S7gE=*>mMVCq0YVL.YYcK,Y=`@\X=h:G[ec2^3h=YF;C/g`2VT*c?qXpKj]]U?cVq
	fC?DNujYDHgQ#<&Y/Z3l"'p#PjIED$6^p\*"iPGcpqYD$G(gq`=S8``85)IVK,7k,W='%mZ6N^%W9s
	ZH4Ib4'RZ77*WgpE7az8OZBBY!QNJ
	ASCII85End
end